import os
import pytest
from pathlib import Path
from .utils import get_lqp_input_files, get_all_files, PARENT_DIR


def get_base_filename(filepath):
    """Get the base filename without extension from a file path."""
    return Path(filepath).stem


def check_output_files_have_corresponding_inputs():
    """Check that all output files have corresponding input files."""
    input_files = get_lqp_input_files()
    input_basenames = {get_base_filename(f) for f in input_files}
    
    missing_inputs = []
    
    output_files = [
        ("lqp_output", ".lqp"),
        ("lqp_debug_output", ".lqp"),
        ("lqp_pretty_output", ".lqp"),
        ("test_files/bin_output", ".bin"),
    ]

    for directory, file_extension in output_files:
        for output_file in get_all_files(PARENT_DIR / directory, file_extension):
            base_name = get_base_filename(output_file)
            if base_name not in input_basenames:
                missing_inputs.append(f"{directory}/{Path(output_file).name} -> missing input {base_name}.lqp")

    return missing_inputs


def test_all_output_files_have_corresponding_inputs():
    """Test that all output files have corresponding input files."""
    missing_inputs = check_output_files_have_corresponding_inputs()
    
    if missing_inputs:
        error_message = "Found output files without corresponding input files:\n" + "\n".join(missing_inputs)
        pytest.fail(error_message)
    


import pytest

from lqp import ir, emit
from lqp import print as lqp_print
from lqp.proto.v1 import logic_pb2

from decimal import Decimal

def test_print_decimal_value():
    d = ir.DecimalValue(precision=18, scale=6, value=Decimal("123456789.123456"), meta=None)
    s = lqp_print.to_str(d, 0)
    assert s == "123456789.123456d18"

    d = ir.DecimalValue(precision=10, scale=2, value=Decimal("0.000123456"), meta=None)
    s = lqp_print.to_str(d, 0)
    assert s == "0.00d10"

    d = ir.DecimalValue(precision=18, scale=6, value=Decimal("0.000123456"), meta=None)
    s = lqp_print.to_str(d, 0)
    assert s == "0.000123d18"

    d = ir.DecimalValue(precision=18, scale=6, value=Decimal("123"), meta=None)
    s = lqp_print.to_str(d, 0)
    assert s == "123.000000d18"

    d = ir.DecimalValue(precision=18, scale=6, value=Decimal(4.4), meta=None)
    s = lqp_print.to_str(d, 0)
    assert s == "4.400000d18"

def test_convert_decimal_value():
    d = ir.DecimalValue(precision=18, scale=6, value=Decimal("123456789.123456"), meta=None)
    v = ir.Value(value=d, meta=None)
    converted = emit.convert_value(v)
    assert converted == logic_pb2.Value(decimal_value=logic_pb2.DecimalValue(
        precision=18,
        scale=6,
        value=logic_pb2.Int128Value(low=123456789123456, high=0)
    ))

    # -4.5 is represented by the digits 4, 5 and a sign of 1
    d = ir.DecimalValue(precision=10, scale=2, value=Decimal(-4.5), meta=None)
    v = ir.Value(value=d, meta=None)
    converted = emit.convert_value(v)
    assert converted == logic_pb2.Value(decimal_value=logic_pb2.DecimalValue(
        precision=10,
        scale=2,
        value=logic_pb2.Int128Value(low=18446744073709551166, high=18446744073709551615)
    ))

    # 4.5 is represented by the digits 4, 5
    d = ir.DecimalValue(precision=10, scale=2, value=Decimal(4.5), meta=None)
    v = ir.Value(value=d, meta=None)
    converted = emit.convert_value(v)
    assert converted == logic_pb2.Value(decimal_value=logic_pb2.DecimalValue(
        precision=10,
        scale=2,
        value=logic_pb2.Int128Value(low=450, high=0)
    ))

    # 4.4 is represented by the digits 4, 4, 0, 0, [43 others], 2, 5
    d = ir.DecimalValue(precision=10, scale=2, value=Decimal(4.4), meta=None)
    v = ir.Value(value=d, meta=None)
    converted = emit.convert_value(v)
    assert converted == logic_pb2.Value(decimal_value=logic_pb2.DecimalValue(
        precision=10,
        scale=2,
        value=logic_pb2.Int128Value(low=440, high=0)
    ))



import os
import re
import pytest
import sys
import lqp.ir as ir
from pathlib import Path
from lqp.parser import parse_lqp
from lqp.emit import ir_to_proto
from lqp.validator import ValidationError, validate_lqp
from pytest_snapshot.plugin import Snapshot
from .utils import get_lqp_input_files

@pytest.mark.parametrize("input_file", get_lqp_input_files())
def test_parse_lqp(snapshot: Snapshot, input_file):
    """Test that each input file can be successfully parsed and matches its binary snapshot"""
    try:
        with open(input_file, "r") as f:
            content = f.read()

        # Parse the file and convert to protobuf
        parsed_lqp = parse_lqp(input_file, content)
        assert parsed_lqp is not None, f"Failed to parse {input_file}"
        proto_result = ir_to_proto(parsed_lqp)
        assert proto_result is not None, f"Failed to convert IR to Proto for {input_file}"
        binary_output = proto_result.SerializeToString()
        snapshot.snapshot_dir = Path(__file__).parent / "test_files" / "bin_output"
        snapshot_filename = os.path.basename(input_file).replace(".lqp", ".bin")
        snapshot.assert_match(binary_output, snapshot_filename)
        print(f"Successfully parsed and snapshotted {input_file}")

    except Exception as e:
        pytest.fail(f"Failed checking {input_file}: {str(e)}")

@pytest.mark.parametrize("input_file", get_lqp_input_files())
def test_validate_lqp_inputs(input_file):
    try:
        with open(input_file, "r") as f:
            content = f.read()
        parsed_lqp = parse_lqp(input_file, content)
        if not(isinstance(parsed_lqp, ir.Transaction)): return
        validate_lqp(parsed_lqp)
    except Exception as e:
        pytest.fail(f"Failed validating {input_file}: {str(e)}")

VALIDATOR_DIR = Path(__file__).parent / "validator"

def test_valid_validator_files():
    for validator_file in VALIDATOR_DIR.glob("valid_*.lqp"):
        with open(validator_file, "r") as f:
            content = f.read()
        try:
            result = parse_lqp(validator_file, content)
            assert result is not None, f"Failed to parse {validator_file}"
            assert isinstance(result, ir.Transaction), f"{validator_file} does not contain a transaction"
            validate_lqp(result)
            print(f"Successfully validated {validator_file}")
        except Exception as e:
            pytest.fail(f"Failed to parse valid validator file {validator_file}: {str(e)}")

def extract_expected_error(file_path):
    with open(file_path, "r") as f:
        content = f.read()
    error_match = re.search(r';;\s*ERROR:\s*(.+)(?:\n|\r\n?)', content)
    if error_match:
        return error_match.group(1).strip()
    return None

@pytest.mark.parametrize("validator_file", [f for f in os.listdir(VALIDATOR_DIR) if f.startswith("fail_")])
def test_validator_failure_files(validator_file):
    file_path = VALIDATOR_DIR / validator_file
    expected_error = extract_expected_error(file_path)
    if not expected_error:
        pytest.skip(f"No expected error comment found in {validator_file}")
        return
    with open(file_path, "r") as f:
        content = f.read()
    result = parse_lqp(validator_file, content)
    assert isinstance(result, ir.Transaction), f"{validator_file} does not contain a transaction"
    with pytest.raises(ValidationError) as exc_info:
        validate_lqp(result)
    error_message = str(exc_info.value)
    assert expected_error in error_message, f"Expected '{expected_error}' in error message: {error_message}"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: relationalai/lqp/v1/transactions.proto
# Protobuf Python Version: 4.25.6
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from lqp.proto.v1 import fragments_pb2 as relationalai_dot_lqp_dot_v1_dot_fragments__pb2
from lqp.proto.v1 import logic_pb2 as relationalai_dot_lqp_dot_v1_dot_logic__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&relationalai/lqp/v1/transactions.proto\x12\x13relationalai.lqp.v1\x1a#relationalai/lqp/v1/fragments.proto\x1a\x1frelationalai/lqp/v1/logic.proto\"\x7f\n\x0bTransaction\x12\x32\n\x06\x65pochs\x18\x01 \x03(\x0b\x32\x1a.relationalai.lqp.v1.EpochR\x06\x65pochs\x12<\n\tconfigure\x18\x02 \x01(\x0b\x32\x1e.relationalai.lqp.v1.ConfigureR\tconfigure\"w\n\tConfigure\x12+\n\x11semantics_version\x18\x01 \x01(\x03R\x10semanticsVersion\x12=\n\nivm_config\x18\x02 \x01(\x0b\x32\x1e.relationalai.lqp.v1.IVMConfigR\tivmConfig\"H\n\tIVMConfig\x12;\n\x05level\x18\x01 \x01(\x0e\x32%.relationalai.lqp.v1.MaintenanceLevelR\x05level\"l\n\x05\x45poch\x12\x32\n\x06writes\x18\x01 \x03(\x0b\x32\x1a.relationalai.lqp.v1.WriteR\x06writes\x12/\n\x05reads\x18\x02 \x03(\x0b\x32\x19.relationalai.lqp.v1.ReadR\x05reads\"\xc3\x01\n\x05Write\x12\x35\n\x06\x64\x65\x66ine\x18\x01 \x01(\x0b\x32\x1b.relationalai.lqp.v1.DefineH\x00R\x06\x64\x65\x66ine\x12;\n\x08undefine\x18\x02 \x01(\x0b\x32\x1d.relationalai.lqp.v1.UndefineH\x00R\x08undefine\x12\x38\n\x07\x63ontext\x18\x03 \x01(\x0b\x32\x1c.relationalai.lqp.v1.ContextH\x00R\x07\x63ontextB\x0c\n\nwrite_type\"C\n\x06\x44\x65\x66ine\x12\x39\n\x08\x66ragment\x18\x01 \x01(\x0b\x32\x1d.relationalai.lqp.v1.FragmentR\x08\x66ragment\"L\n\x08Undefine\x12@\n\x0b\x66ragment_id\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.FragmentIdR\nfragmentId\"H\n\x07\x43ontext\x12=\n\trelations\x18\x01 \x03(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\trelations\"\xc4\x04\n\x0f\x45xportCSVConfig\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\x12G\n\x0c\x64\x61ta_columns\x18\x02 \x03(\x0b\x32$.relationalai.lqp.v1.ExportCSVColumnR\x0b\x64\x61taColumns\x12*\n\x0epartition_size\x18\x03 \x01(\x03H\x00R\rpartitionSize\x88\x01\x01\x12%\n\x0b\x63ompression\x18\x04 \x01(\tH\x01R\x0b\x63ompression\x88\x01\x01\x12/\n\x11syntax_header_row\x18\x05 \x01(\x08H\x02R\x0fsyntaxHeaderRow\x88\x01\x01\x12\x37\n\x15syntax_missing_string\x18\x06 \x01(\tH\x03R\x13syntaxMissingString\x88\x01\x01\x12&\n\x0csyntax_delim\x18\x07 \x01(\tH\x04R\x0bsyntaxDelim\x88\x01\x01\x12.\n\x10syntax_quotechar\x18\x08 \x01(\tH\x05R\x0fsyntaxQuotechar\x88\x01\x01\x12\x30\n\x11syntax_escapechar\x18\t \x01(\tH\x06R\x10syntaxEscapechar\x88\x01\x01\x42\x11\n\x0f_partition_sizeB\x0e\n\x0c_compressionB\x14\n\x12_syntax_header_rowB\x18\n\x16_syntax_missing_stringB\x0f\n\r_syntax_delimB\x13\n\x11_syntax_quotecharB\x14\n\x12_syntax_escapechar\"t\n\x0f\x45xportCSVColumn\x12\x1f\n\x0b\x63olumn_name\x18\x01 \x01(\tR\ncolumnName\x12@\n\x0b\x63olumn_data\x18\x02 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\ncolumnData\"\xa4\x02\n\x04Read\x12\x35\n\x06\x64\x65mand\x18\x01 \x01(\x0b\x32\x1b.relationalai.lqp.v1.DemandH\x00R\x06\x64\x65mand\x12\x35\n\x06output\x18\x02 \x01(\x0b\x32\x1b.relationalai.lqp.v1.OutputH\x00R\x06output\x12\x36\n\x07what_if\x18\x03 \x01(\x0b\x32\x1b.relationalai.lqp.v1.WhatIfH\x00R\x06whatIf\x12\x32\n\x05\x61\x62ort\x18\x04 \x01(\x0b\x32\x1a.relationalai.lqp.v1.AbortH\x00R\x05\x61\x62ort\x12\x35\n\x06\x65xport\x18\x05 \x01(\x0b\x32\x1b.relationalai.lqp.v1.ExportH\x00R\x06\x65xportB\x0b\n\tread_type\"J\n\x06\x44\x65mand\x12@\n\x0brelation_id\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\nrelationId\"^\n\x06Output\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12@\n\x0brelation_id\x18\x02 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\nrelationId\"`\n\x06\x45xport\x12\x45\n\ncsv_config\x18\x01 \x01(\x0b\x32$.relationalai.lqp.v1.ExportCSVConfigH\x00R\tcsvConfigB\x0f\n\rexport_config\"R\n\x06WhatIf\x12\x16\n\x06\x62ranch\x18\x01 \x01(\tR\x06\x62ranch\x12\x30\n\x05\x65poch\x18\x02 \x01(\x0b\x32\x1a.relationalai.lqp.v1.EpochR\x05\x65poch\"]\n\x05\x41\x62ort\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12@\n\x0brelation_id\x18\x02 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\nrelationId*\x87\x01\n\x10MaintenanceLevel\x12!\n\x1dMAINTENANCE_LEVEL_UNSPECIFIED\x10\x00\x12\x19\n\x15MAINTENANCE_LEVEL_OFF\x10\x01\x12\x1a\n\x16MAINTENANCE_LEVEL_AUTO\x10\x02\x12\x19\n\x15MAINTENANCE_LEVEL_ALL\x10\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'relationalai.lqp.v1.transactions_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_MAINTENANCELEVEL']._serialized_start=2432
  _globals['_MAINTENANCELEVEL']._serialized_end=2567
  _globals['_TRANSACTION']._serialized_start=133
  _globals['_TRANSACTION']._serialized_end=260
  _globals['_CONFIGURE']._serialized_start=262
  _globals['_CONFIGURE']._serialized_end=381
  _globals['_IVMCONFIG']._serialized_start=383
  _globals['_IVMCONFIG']._serialized_end=455
  _globals['_EPOCH']._serialized_start=457
  _globals['_EPOCH']._serialized_end=565
  _globals['_WRITE']._serialized_start=568
  _globals['_WRITE']._serialized_end=763
  _globals['_DEFINE']._serialized_start=765
  _globals['_DEFINE']._serialized_end=832
  _globals['_UNDEFINE']._serialized_start=834
  _globals['_UNDEFINE']._serialized_end=910
  _globals['_CONTEXT']._serialized_start=912
  _globals['_CONTEXT']._serialized_end=984
  _globals['_EXPORTCSVCONFIG']._serialized_start=987
  _globals['_EXPORTCSVCONFIG']._serialized_end=1567
  _globals['_EXPORTCSVCOLUMN']._serialized_start=1569
  _globals['_EXPORTCSVCOLUMN']._serialized_end=1685
  _globals['_READ']._serialized_start=1688
  _globals['_READ']._serialized_end=1980
  _globals['_DEMAND']._serialized_start=1982
  _globals['_DEMAND']._serialized_end=2056
  _globals['_OUTPUT']._serialized_start=2058
  _globals['_OUTPUT']._serialized_end=2152
  _globals['_EXPORT']._serialized_start=2154
  _globals['_EXPORT']._serialized_end=2250
  _globals['_WHATIF']._serialized_start=2252
  _globals['_WHATIF']._serialized_end=2334
  _globals['_ABORT']._serialized_start=2336
  _globals['_ABORT']._serialized_end=2429
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Declaration(_message.Message):
    __slots__ = ("algorithm",)
    DEF_FIELD_NUMBER: _ClassVar[int]
    ALGORITHM_FIELD_NUMBER: _ClassVar[int]
    algorithm: Algorithm
    def __init__(self, algorithm: _Optional[_Union[Algorithm, _Mapping]] = ..., **kwargs) -> None: ...

class Def(_message.Message):
    __slots__ = ("name", "body", "attrs")
    NAME_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    ATTRS_FIELD_NUMBER: _ClassVar[int]
    name: RelationId
    body: Abstraction
    attrs: _containers.RepeatedCompositeFieldContainer[Attribute]
    def __init__(self, name: _Optional[_Union[RelationId, _Mapping]] = ..., body: _Optional[_Union[Abstraction, _Mapping]] = ..., attrs: _Optional[_Iterable[_Union[Attribute, _Mapping]]] = ...) -> None: ...

class Algorithm(_message.Message):
    __slots__ = ("body",)
    GLOBAL_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    body: Script
    def __init__(self, body: _Optional[_Union[Script, _Mapping]] = ..., **kwargs) -> None: ...

class Script(_message.Message):
    __slots__ = ("constructs",)
    CONSTRUCTS_FIELD_NUMBER: _ClassVar[int]
    constructs: _containers.RepeatedCompositeFieldContainer[Construct]
    def __init__(self, constructs: _Optional[_Iterable[_Union[Construct, _Mapping]]] = ...) -> None: ...

class Construct(_message.Message):
    __slots__ = ("loop", "instruction")
    LOOP_FIELD_NUMBER: _ClassVar[int]
    INSTRUCTION_FIELD_NUMBER: _ClassVar[int]
    loop: Loop
    instruction: Instruction
    def __init__(self, loop: _Optional[_Union[Loop, _Mapping]] = ..., instruction: _Optional[_Union[Instruction, _Mapping]] = ...) -> None: ...

class Loop(_message.Message):
    __slots__ = ("init", "body")
    INIT_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    init: _containers.RepeatedCompositeFieldContainer[Instruction]
    body: Script
    def __init__(self, init: _Optional[_Iterable[_Union[Instruction, _Mapping]]] = ..., body: _Optional[_Union[Script, _Mapping]] = ...) -> None: ...

class Instruction(_message.Message):
    __slots__ = ("assign", "upsert", "copy", "monoid_def", "monus_def")
    ASSIGN_FIELD_NUMBER: _ClassVar[int]
    UPSERT_FIELD_NUMBER: _ClassVar[int]
    BREAK_FIELD_NUMBER: _ClassVar[int]
    COPY_FIELD_NUMBER: _ClassVar[int]
    MONOID_DEF_FIELD_NUMBER: _ClassVar[int]
    MONUS_DEF_FIELD_NUMBER: _ClassVar[int]
    assign: Assign
    upsert: Upsert
    copy: Copy
    monoid_def: MonoidDef
    monus_def: MonusDef
    def __init__(self, assign: _Optional[_Union[Assign, _Mapping]] = ..., upsert: _Optional[_Union[Upsert, _Mapping]] = ..., copy: _Optional[_Union[Copy, _Mapping]] = ..., monoid_def: _Optional[_Union[MonoidDef, _Mapping]] = ..., monus_def: _Optional[_Union[MonusDef, _Mapping]] = ..., **kwargs) -> None: ...

class Assign(_message.Message):
    __slots__ = ("name", "body", "attrs")
    NAME_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    ATTRS_FIELD_NUMBER: _ClassVar[int]
    name: RelationId
    body: Abstraction
    attrs: _containers.RepeatedCompositeFieldContainer[Attribute]
    def __init__(self, name: _Optional[_Union[RelationId, _Mapping]] = ..., body: _Optional[_Union[Abstraction, _Mapping]] = ..., attrs: _Optional[_Iterable[_Union[Attribute, _Mapping]]] = ...) -> None: ...

class Upsert(_message.Message):
    __slots__ = ("name", "body", "attrs")
    NAME_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    ATTRS_FIELD_NUMBER: _ClassVar[int]
    name: RelationId
    body: Abstraction
    attrs: _containers.RepeatedCompositeFieldContainer[Attribute]
    def __init__(self, name: _Optional[_Union[RelationId, _Mapping]] = ..., body: _Optional[_Union[Abstraction, _Mapping]] = ..., attrs: _Optional[_Iterable[_Union[Attribute, _Mapping]]] = ...) -> None: ...

class Break(_message.Message):
    __slots__ = ("name", "body", "attrs")
    NAME_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    ATTRS_FIELD_NUMBER: _ClassVar[int]
    name: RelationId
    body: Abstraction
    attrs: _containers.RepeatedCompositeFieldContainer[Attribute]
    def __init__(self, name: _Optional[_Union[RelationId, _Mapping]] = ..., body: _Optional[_Union[Abstraction, _Mapping]] = ..., attrs: _Optional[_Iterable[_Union[Attribute, _Mapping]]] = ...) -> None: ...

class Copy(_message.Message):
    __slots__ = ("name", "body", "attrs")
    NAME_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    ATTRS_FIELD_NUMBER: _ClassVar[int]
    name: RelationId
    body: Abstraction
    attrs: _containers.RepeatedCompositeFieldContainer[Attribute]
    def __init__(self, name: _Optional[_Union[RelationId, _Mapping]] = ..., body: _Optional[_Union[Abstraction, _Mapping]] = ..., attrs: _Optional[_Iterable[_Union[Attribute, _Mapping]]] = ...) -> None: ...

class MonoidDef(_message.Message):
    __slots__ = ("monoid", "name", "body", "attrs")
    MONOID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    ATTRS_FIELD_NUMBER: _ClassVar[int]
    monoid: Monoid
    name: RelationId
    body: Abstraction
    attrs: _containers.RepeatedCompositeFieldContainer[Attribute]
    def __init__(self, monoid: _Optional[_Union[Monoid, _Mapping]] = ..., name: _Optional[_Union[RelationId, _Mapping]] = ..., body: _Optional[_Union[Abstraction, _Mapping]] = ..., attrs: _Optional[_Iterable[_Union[Attribute, _Mapping]]] = ...) -> None: ...

class MonusDef(_message.Message):
    __slots__ = ("monoid", "name", "body", "attrs")
    MONOID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    ATTRS_FIELD_NUMBER: _ClassVar[int]
    monoid: Monoid
    name: RelationId
    body: Abstraction
    attrs: _containers.RepeatedCompositeFieldContainer[Attribute]
    def __init__(self, monoid: _Optional[_Union[Monoid, _Mapping]] = ..., name: _Optional[_Union[RelationId, _Mapping]] = ..., body: _Optional[_Union[Abstraction, _Mapping]] = ..., attrs: _Optional[_Iterable[_Union[Attribute, _Mapping]]] = ...) -> None: ...

class Monoid(_message.Message):
    __slots__ = ("or_monoid", "min_monoid", "max_monoid", "sum_monoid")
    OR_MONOID_FIELD_NUMBER: _ClassVar[int]
    MIN_MONOID_FIELD_NUMBER: _ClassVar[int]
    MAX_MONOID_FIELD_NUMBER: _ClassVar[int]
    SUM_MONOID_FIELD_NUMBER: _ClassVar[int]
    or_monoid: OrMonoid
    min_monoid: MinMonoid
    max_monoid: MaxMonoid
    sum_monoid: SumMonoid
    def __init__(self, or_monoid: _Optional[_Union[OrMonoid, _Mapping]] = ..., min_monoid: _Optional[_Union[MinMonoid, _Mapping]] = ..., max_monoid: _Optional[_Union[MaxMonoid, _Mapping]] = ..., sum_monoid: _Optional[_Union[SumMonoid, _Mapping]] = ...) -> None: ...

class OrMonoid(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class MinMonoid(_message.Message):
    __slots__ = ("type",)
    TYPE_FIELD_NUMBER: _ClassVar[int]
    type: Type
    def __init__(self, type: _Optional[_Union[Type, _Mapping]] = ...) -> None: ...

class MaxMonoid(_message.Message):
    __slots__ = ("type",)
    TYPE_FIELD_NUMBER: _ClassVar[int]
    type: Type
    def __init__(self, type: _Optional[_Union[Type, _Mapping]] = ...) -> None: ...

class SumMonoid(_message.Message):
    __slots__ = ("type",)
    TYPE_FIELD_NUMBER: _ClassVar[int]
    type: Type
    def __init__(self, type: _Optional[_Union[Type, _Mapping]] = ...) -> None: ...

class Binding(_message.Message):
    __slots__ = ("var", "type")
    VAR_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    var: Var
    type: Type
    def __init__(self, var: _Optional[_Union[Var, _Mapping]] = ..., type: _Optional[_Union[Type, _Mapping]] = ...) -> None: ...

class Abstraction(_message.Message):
    __slots__ = ("vars", "value")
    VARS_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    vars: _containers.RepeatedCompositeFieldContainer[Binding]
    value: Formula
    def __init__(self, vars: _Optional[_Iterable[_Union[Binding, _Mapping]]] = ..., value: _Optional[_Union[Formula, _Mapping]] = ...) -> None: ...

class Formula(_message.Message):
    __slots__ = ("exists", "reduce", "conjunction", "disjunction", "ffi", "atom", "pragma", "primitive", "rel_atom", "cast")
    EXISTS_FIELD_NUMBER: _ClassVar[int]
    REDUCE_FIELD_NUMBER: _ClassVar[int]
    CONJUNCTION_FIELD_NUMBER: _ClassVar[int]
    DISJUNCTION_FIELD_NUMBER: _ClassVar[int]
    NOT_FIELD_NUMBER: _ClassVar[int]
    FFI_FIELD_NUMBER: _ClassVar[int]
    ATOM_FIELD_NUMBER: _ClassVar[int]
    PRAGMA_FIELD_NUMBER: _ClassVar[int]
    PRIMITIVE_FIELD_NUMBER: _ClassVar[int]
    REL_ATOM_FIELD_NUMBER: _ClassVar[int]
    CAST_FIELD_NUMBER: _ClassVar[int]
    exists: Exists
    reduce: Reduce
    conjunction: Conjunction
    disjunction: Disjunction
    ffi: FFI
    atom: Atom
    pragma: Pragma
    primitive: Primitive
    rel_atom: RelAtom
    cast: Cast
    def __init__(self, exists: _Optional[_Union[Exists, _Mapping]] = ..., reduce: _Optional[_Union[Reduce, _Mapping]] = ..., conjunction: _Optional[_Union[Conjunction, _Mapping]] = ..., disjunction: _Optional[_Union[Disjunction, _Mapping]] = ..., ffi: _Optional[_Union[FFI, _Mapping]] = ..., atom: _Optional[_Union[Atom, _Mapping]] = ..., pragma: _Optional[_Union[Pragma, _Mapping]] = ..., primitive: _Optional[_Union[Primitive, _Mapping]] = ..., rel_atom: _Optional[_Union[RelAtom, _Mapping]] = ..., cast: _Optional[_Union[Cast, _Mapping]] = ..., **kwargs) -> None: ...

class Exists(_message.Message):
    __slots__ = ("body",)
    BODY_FIELD_NUMBER: _ClassVar[int]
    body: Abstraction
    def __init__(self, body: _Optional[_Union[Abstraction, _Mapping]] = ...) -> None: ...

class Reduce(_message.Message):
    __slots__ = ("op", "body", "terms")
    OP_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    op: Abstraction
    body: Abstraction
    terms: _containers.RepeatedCompositeFieldContainer[Term]
    def __init__(self, op: _Optional[_Union[Abstraction, _Mapping]] = ..., body: _Optional[_Union[Abstraction, _Mapping]] = ..., terms: _Optional[_Iterable[_Union[Term, _Mapping]]] = ...) -> None: ...

class Conjunction(_message.Message):
    __slots__ = ("args",)
    ARGS_FIELD_NUMBER: _ClassVar[int]
    args: _containers.RepeatedCompositeFieldContainer[Formula]
    def __init__(self, args: _Optional[_Iterable[_Union[Formula, _Mapping]]] = ...) -> None: ...

class Disjunction(_message.Message):
    __slots__ = ("args",)
    ARGS_FIELD_NUMBER: _ClassVar[int]
    args: _containers.RepeatedCompositeFieldContainer[Formula]
    def __init__(self, args: _Optional[_Iterable[_Union[Formula, _Mapping]]] = ...) -> None: ...

class Not(_message.Message):
    __slots__ = ("arg",)
    ARG_FIELD_NUMBER: _ClassVar[int]
    arg: Formula
    def __init__(self, arg: _Optional[_Union[Formula, _Mapping]] = ...) -> None: ...

class FFI(_message.Message):
    __slots__ = ("name", "args", "terms")
    NAME_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    name: str
    args: _containers.RepeatedCompositeFieldContainer[Abstraction]
    terms: _containers.RepeatedCompositeFieldContainer[Term]
    def __init__(self, name: _Optional[str] = ..., args: _Optional[_Iterable[_Union[Abstraction, _Mapping]]] = ..., terms: _Optional[_Iterable[_Union[Term, _Mapping]]] = ...) -> None: ...

class Atom(_message.Message):
    __slots__ = ("name", "terms")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    name: RelationId
    terms: _containers.RepeatedCompositeFieldContainer[Term]
    def __init__(self, name: _Optional[_Union[RelationId, _Mapping]] = ..., terms: _Optional[_Iterable[_Union[Term, _Mapping]]] = ...) -> None: ...

class Pragma(_message.Message):
    __slots__ = ("name", "terms")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    name: str
    terms: _containers.RepeatedCompositeFieldContainer[Term]
    def __init__(self, name: _Optional[str] = ..., terms: _Optional[_Iterable[_Union[Term, _Mapping]]] = ...) -> None: ...

class Primitive(_message.Message):
    __slots__ = ("name", "terms")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    name: str
    terms: _containers.RepeatedCompositeFieldContainer[RelTerm]
    def __init__(self, name: _Optional[str] = ..., terms: _Optional[_Iterable[_Union[RelTerm, _Mapping]]] = ...) -> None: ...

class RelAtom(_message.Message):
    __slots__ = ("name", "terms")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TERMS_FIELD_NUMBER: _ClassVar[int]
    name: str
    terms: _containers.RepeatedCompositeFieldContainer[RelTerm]
    def __init__(self, name: _Optional[str] = ..., terms: _Optional[_Iterable[_Union[RelTerm, _Mapping]]] = ...) -> None: ...

class Cast(_message.Message):
    __slots__ = ("input", "result")
    INPUT_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    input: Term
    result: Term
    def __init__(self, input: _Optional[_Union[Term, _Mapping]] = ..., result: _Optional[_Union[Term, _Mapping]] = ...) -> None: ...

class RelTerm(_message.Message):
    __slots__ = ("specialized_value", "term")
    SPECIALIZED_VALUE_FIELD_NUMBER: _ClassVar[int]
    TERM_FIELD_NUMBER: _ClassVar[int]
    specialized_value: Value
    term: Term
    def __init__(self, specialized_value: _Optional[_Union[Value, _Mapping]] = ..., term: _Optional[_Union[Term, _Mapping]] = ...) -> None: ...

class Term(_message.Message):
    __slots__ = ("var", "constant")
    VAR_FIELD_NUMBER: _ClassVar[int]
    CONSTANT_FIELD_NUMBER: _ClassVar[int]
    var: Var
    constant: Value
    def __init__(self, var: _Optional[_Union[Var, _Mapping]] = ..., constant: _Optional[_Union[Value, _Mapping]] = ...) -> None: ...

class Var(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class Attribute(_message.Message):
    __slots__ = ("name", "args")
    NAME_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    name: str
    args: _containers.RepeatedCompositeFieldContainer[Value]
    def __init__(self, name: _Optional[str] = ..., args: _Optional[_Iterable[_Union[Value, _Mapping]]] = ...) -> None: ...

class RelationId(_message.Message):
    __slots__ = ("id_low", "id_high")
    ID_LOW_FIELD_NUMBER: _ClassVar[int]
    ID_HIGH_FIELD_NUMBER: _ClassVar[int]
    id_low: int
    id_high: int
    def __init__(self, id_low: _Optional[int] = ..., id_high: _Optional[int] = ...) -> None: ...

class Type(_message.Message):
    __slots__ = ("unspecified_type", "string_type", "int_type", "float_type", "uint128_type", "int128_type", "date_type", "datetime_type", "missing_type", "decimal_type", "boolean_type")
    UNSPECIFIED_TYPE_FIELD_NUMBER: _ClassVar[int]
    STRING_TYPE_FIELD_NUMBER: _ClassVar[int]
    INT_TYPE_FIELD_NUMBER: _ClassVar[int]
    FLOAT_TYPE_FIELD_NUMBER: _ClassVar[int]
    UINT128_TYPE_FIELD_NUMBER: _ClassVar[int]
    INT128_TYPE_FIELD_NUMBER: _ClassVar[int]
    DATE_TYPE_FIELD_NUMBER: _ClassVar[int]
    DATETIME_TYPE_FIELD_NUMBER: _ClassVar[int]
    MISSING_TYPE_FIELD_NUMBER: _ClassVar[int]
    DECIMAL_TYPE_FIELD_NUMBER: _ClassVar[int]
    BOOLEAN_TYPE_FIELD_NUMBER: _ClassVar[int]
    unspecified_type: UnspecifiedType
    string_type: StringType
    int_type: IntType
    float_type: FloatType
    uint128_type: UInt128Type
    int128_type: Int128Type
    date_type: DateType
    datetime_type: DateTimeType
    missing_type: MissingType
    decimal_type: DecimalType
    boolean_type: BooleanType
    def __init__(self, unspecified_type: _Optional[_Union[UnspecifiedType, _Mapping]] = ..., string_type: _Optional[_Union[StringType, _Mapping]] = ..., int_type: _Optional[_Union[IntType, _Mapping]] = ..., float_type: _Optional[_Union[FloatType, _Mapping]] = ..., uint128_type: _Optional[_Union[UInt128Type, _Mapping]] = ..., int128_type: _Optional[_Union[Int128Type, _Mapping]] = ..., date_type: _Optional[_Union[DateType, _Mapping]] = ..., datetime_type: _Optional[_Union[DateTimeType, _Mapping]] = ..., missing_type: _Optional[_Union[MissingType, _Mapping]] = ..., decimal_type: _Optional[_Union[DecimalType, _Mapping]] = ..., boolean_type: _Optional[_Union[BooleanType, _Mapping]] = ...) -> None: ...

class UnspecifiedType(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class StringType(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class IntType(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class FloatType(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class UInt128Type(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class Int128Type(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class DateType(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class DateTimeType(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class MissingType(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class DecimalType(_message.Message):
    __slots__ = ("precision", "scale")
    PRECISION_FIELD_NUMBER: _ClassVar[int]
    SCALE_FIELD_NUMBER: _ClassVar[int]
    precision: int
    scale: int
    def __init__(self, precision: _Optional[int] = ..., scale: _Optional[int] = ...) -> None: ...

class BooleanType(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class Value(_message.Message):
    __slots__ = ("string_value", "int_value", "float_value", "uint128_value", "int128_value", "missing_value", "date_value", "datetime_value", "decimal_value", "boolean_value")
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
    UINT128_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT128_VALUE_FIELD_NUMBER: _ClassVar[int]
    MISSING_VALUE_FIELD_NUMBER: _ClassVar[int]
    DATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    DATETIME_VALUE_FIELD_NUMBER: _ClassVar[int]
    DECIMAL_VALUE_FIELD_NUMBER: _ClassVar[int]
    BOOLEAN_VALUE_FIELD_NUMBER: _ClassVar[int]
    string_value: str
    int_value: int
    float_value: float
    uint128_value: UInt128Value
    int128_value: Int128Value
    missing_value: MissingValue
    date_value: DateValue
    datetime_value: DateTimeValue
    decimal_value: DecimalValue
    boolean_value: bool
    def __init__(self, string_value: _Optional[str] = ..., int_value: _Optional[int] = ..., float_value: _Optional[float] = ..., uint128_value: _Optional[_Union[UInt128Value, _Mapping]] = ..., int128_value: _Optional[_Union[Int128Value, _Mapping]] = ..., missing_value: _Optional[_Union[MissingValue, _Mapping]] = ..., date_value: _Optional[_Union[DateValue, _Mapping]] = ..., datetime_value: _Optional[_Union[DateTimeValue, _Mapping]] = ..., decimal_value: _Optional[_Union[DecimalValue, _Mapping]] = ..., boolean_value: bool = ...) -> None: ...

class UInt128Value(_message.Message):
    __slots__ = ("low", "high")
    LOW_FIELD_NUMBER: _ClassVar[int]
    HIGH_FIELD_NUMBER: _ClassVar[int]
    low: int
    high: int
    def __init__(self, low: _Optional[int] = ..., high: _Optional[int] = ...) -> None: ...

class Int128Value(_message.Message):
    __slots__ = ("low", "high")
    LOW_FIELD_NUMBER: _ClassVar[int]
    HIGH_FIELD_NUMBER: _ClassVar[int]
    low: int
    high: int
    def __init__(self, low: _Optional[int] = ..., high: _Optional[int] = ...) -> None: ...

class MissingValue(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class DateValue(_message.Message):
    __slots__ = ("year", "month", "day")
    YEAR_FIELD_NUMBER: _ClassVar[int]
    MONTH_FIELD_NUMBER: _ClassVar[int]
    DAY_FIELD_NUMBER: _ClassVar[int]
    year: int
    month: int
    day: int
    def __init__(self, year: _Optional[int] = ..., month: _Optional[int] = ..., day: _Optional[int] = ...) -> None: ...

class DateTimeValue(_message.Message):
    __slots__ = ("year", "month", "day", "hour", "minute", "second", "microsecond")
    YEAR_FIELD_NUMBER: _ClassVar[int]
    MONTH_FIELD_NUMBER: _ClassVar[int]
    DAY_FIELD_NUMBER: _ClassVar[int]
    HOUR_FIELD_NUMBER: _ClassVar[int]
    MINUTE_FIELD_NUMBER: _ClassVar[int]
    SECOND_FIELD_NUMBER: _ClassVar[int]
    MICROSECOND_FIELD_NUMBER: _ClassVar[int]
    year: int
    month: int
    day: int
    hour: int
    minute: int
    second: int
    microsecond: int
    def __init__(self, year: _Optional[int] = ..., month: _Optional[int] = ..., day: _Optional[int] = ..., hour: _Optional[int] = ..., minute: _Optional[int] = ..., second: _Optional[int] = ..., microsecond: _Optional[int] = ...) -> None: ...

class DecimalValue(_message.Message):
    __slots__ = ("precision", "scale", "value")
    PRECISION_FIELD_NUMBER: _ClassVar[int]
    SCALE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    precision: int
    scale: int
    value: Int128Value
    def __init__(self, precision: _Optional[int] = ..., scale: _Optional[int] = ..., value: _Optional[_Union[Int128Value, _Mapping]] = ...) -> None: ...



import pytest
import os
import dataclasses
from pytest_snapshot.plugin import Snapshot
from math import isnan

from lqp import parser
from lqp import print as lqp_print
from lqp import ir
from .utils import get_lqp_input_files

def assert_lqp_nodes_equal(obj1, obj2):
    if isinstance(obj1, ir.LqpNode) and isinstance(obj2, ir.LqpNode):
        if obj1.__class__ is not obj2.__class__:
            raise AssertionError(f"Node types differ: {obj1.__class__} vs {obj2.__class__}")
        for field_info in dataclasses.fields(type(obj1)):
            if field_info.name == 'meta': continue
            if field_info.name.startswith('debug_'): continue
            assert_lqp_nodes_equal(getattr(obj1, field_info.name), getattr(obj2, field_info.name))
    elif isinstance(obj1, (list, tuple)) and isinstance(obj2, (list, tuple)):
        if len(obj1) != len(obj2):
            raise AssertionError(f"Sequence lengths differ: {len(obj1)} vs {len(obj2)}")
        for i, (item1, item2) in enumerate(zip(obj1, obj2)):
            assert_lqp_nodes_equal(item1, item2)
    elif isinstance(obj1, float) and isinstance(obj2, float):
        if isnan(obj1) and isnan(obj2):
            return
        if obj1 != obj2:
            raise AssertionError(f"Values differ: {obj1} vs {obj2}")
    elif obj1 != obj2:
        raise AssertionError(f"Values differ: {obj1} vs {obj2}")

@pytest.mark.parametrize("input_file", get_lqp_input_files())
def test_print_snapshot(snapshot, input_file):
    with open(input_file, "r") as f:
        original_lqp_str = f.read()
    parsed_node = parser.parse_lqp(input_file, original_lqp_str)
    options = lqp_print.ugly_config.copy()
    options[str(lqp_print.PrettyOptions.PRINT_DEBUG)] = False
    printed_lqp_str = lqp_print.to_string(parsed_node, options)
    snapshot.snapshot_dir = "tests/lqp_output"
    snapshot.assert_match(printed_lqp_str, os.path.basename(input_file))
    re_parsed_node = parser.parse_lqp("reparsed_output.lqp", printed_lqp_str)
    assert_lqp_nodes_equal(re_parsed_node, parsed_node)

@pytest.mark.parametrize("input_file", get_lqp_input_files())
def test_print_debug_snapshot(snapshot, input_file):
    with open(input_file, "r") as f:
        original_lqp_str = f.read()
    parsed_node = parser.parse_lqp(input_file, original_lqp_str)
    options = lqp_print.ugly_config.copy()
    options[str(lqp_print.PrettyOptions.PRINT_DEBUG)] = True
    printed_lqp_str = lqp_print.to_string(parsed_node, options)
    snapshot.snapshot_dir = "tests/lqp_debug_output"
    snapshot.assert_match(printed_lqp_str, os.path.basename(input_file))
    re_parsed_node = parser.parse_lqp("reparsed_output.lqp", printed_lqp_str)
    assert_lqp_nodes_equal(re_parsed_node, parsed_node)

@pytest.mark.parametrize("input_file", get_lqp_input_files())
def test_print_pretty_snapshot(snapshot, input_file):
    with open(input_file, "r") as f:
        original_lqp_str = f.read()
    parsed_node = parser.parse_lqp(input_file, original_lqp_str)
    printed_lqp_str = lqp_print.to_string(parsed_node, {})
    options = lqp_print.ugly_config.copy()
    options[str(lqp_print.PrettyOptions.PRINT_NAMES)] = True
    options[str(lqp_print.PrettyOptions.PRINT_DEBUG)] = False
    pretty_printed_lqp_str = lqp_print.to_string(parsed_node, options)
    snapshot.snapshot_dir = "tests/lqp_pretty_output"
    snapshot.assert_match(pretty_printed_lqp_str, os.path.basename(input_file))
    re_parsed_node = parser.parse_lqp("reparsed_output.lqp", pretty_printed_lqp_str)
    assert_lqp_nodes_equal(re_parsed_node, parsed_node)
